(* ------------------------------------------------------ *)
(*                   GRA0HRC.MOD                          *)
(*            Grafikbibiothek fr Hercules                *)
(*       (c) 1990  H.Zinner, G.Clarici & TOOLBOX          *)
(* ------------------------------------------------------ *)
IMPLEMENTATION MODULE Graf0Hrc;

FROM SYSTEM     IMPORT ASSEMBLER, ADR;
FROM FileSystem IMPORT File, read, notdone, Lookup,
                       Close, ReadNBytes;
FROM Strings    IMPORT Length;

CONST
  FontDateiName = 'FONT.FON';

VAR
  graph        : ARRAY [1..12] OF CARDINAL;
  txt          : ARRAY [1..12] OF CARDINAL;
  Font         : ARRAY [0..255],[0..13] OF CHAR;
  Zeile        : ARRAY [0..347] OF CARDINAL;
  ZSpalte      : ARRAY [1..80] OF CARDINAL;
  ZVerschieben : ARRAY [1..80] OF CARDINAL;

  PROCEDURE OutPort(PortAdr, OutValue : CARDINAL);
  BEGIN
    ASM
      MOV  AX, OutValue
      MOV  DX, PortAdr
      OUT  DX, AL
    END;
  END OutPort;

  PROCEDURE Peek(Segment, Offset : CARDINAL) : CHAR;
  VAR
    ergebnis : CHAR;
  BEGIN
    ASM
      PUSH ES
      MOV  AX,Segment
      MOV  ES,AX
      MOV  BX,Offset
      MOV  AL,ES:[BX]
      MOV  ergebnis,AL
    END;
    RETURN ergebnis;
  END Peek;

  PROCEDURE FontLesen(Fontname : ARRAY OF CHAR) : BOOLEAN;
  VAR
    f       : File;
    gelesen : CARDINAL;
  BEGIN
    f.mode := read;
    Lookup(f, Fontname, FALSE);
    IF f.res = notdone THEN
      RETURN FALSE;
    END;
    ReadNBytes(f, ADR(Font[0,0]), 3584, gelesen);
    IF gelesen <> 3584 THEN
      RETURN FALSE;
    END;
    Close(f);
    RETURN TRUE;
  END FontLesen;

  PROCEDURE ROMFontLesen(GrafikZeichen : BOOLEAN);
  VAR
    buchstabe : CARDINAL;
    zeile     : CARDINAL;
    BasisSeg  : CARDINAL;
    BasisOff  : CARDINAL;
  BEGIN
    FOR buchstabe := 0 TO 127 DO
      FOR zeile := 0 TO 13 DO
        IF (zeile < 3) OR (zeile > 10) THEN
          Font[buchstabe][zeile] := 0C;
        ELSE
          Font[buchstabe][zeile] :=
              Peek(0FFA6H, 0EH + buchstabe * 8 + zeile - 3);
        END;
      END;
    END;
    BasisSeg := 256 * ORD(Peek(0,7FH)) + ORD(Peek(0,7EH));
    BasisOff := 256 * ORD(Peek(0,7DH)) + ORD(Peek(0,7CH));
    ASM
      MOV  AX,BasisOff
      MOV  BX,BasisSeg
      PUSH AX
      AND  AX,0FFF0H
      MOV  CL,4
      SHR  AX,CL
      ADD  BX,AX
      POP  AX
      AND  AX,0FH
      MOV  BasisSeg,BX
      MOV  BasisOff,AX
    END;
    IF GrafikZeichen  OR
       ((BasisSeg > 0) AND (BasisOff > 0)) THEN
      FOR buchstabe := 128 TO 255 DO
        FOR zeile := 0 TO 13 DO
          IF (zeile < 3) OR (zeile > 10) THEN
            Font[buchstabe][zeile] := 0C;
          ELSE
            Font[buchstabe][zeile]:=
              Peek(BasisSeg,
              BasisOff + (buchstabe - 128) * 8 + zeile - 3);
          END;
        END;
      END;
    END;
  END ROMFontLesen;

  PROCEDURE InitialisiereGrafik;
  VAR
    i, u : CARDINAL;
  BEGIN
    (* Registerwerte initialisieren *)
    (* 1. Grafikmodus *)
    graph[ 1] := 35H;
    graph[ 2] := 2DH;
    graph[ 3] := 2EH;
    graph[ 4] := 07H;
    graph[ 5] := 5BH;
    graph[ 6] := 02H;
    graph[ 7] := 57H;
    graph[ 8] := 57H;
    graph[ 9] := 02H;
    graph[10] := 03H;
    graph[11] := 00H;
    graph[12] := 00H;

    (* 2. Textmodus *)
    txt[ 1] := 61H;
    txt[ 2] := 50H;
    txt[ 3] := 52H;
    txt[ 4] := 0FH;
    txt[ 5] := 19H;
    txt[ 6] := 06H;
    txt[ 7] := 19H;
    txt[ 8] := 19H;
    txt[ 9] := 02H;
    txt[10] := 0DH;
    txt[11] := 0BH;
    txt[12] := 0CH;

    OutPort(03BFH, 3);
    (* beide Seiten lschen *)
    ASM
      PUSH ES
      PUSH DI
      MOV  AX, 0B000H
      MOV  ES, AX
      XOR  DI, DI
      MOV  CX, 8000H
      XOR  AX, AX
      REPZ STOSW
      POP  DI
      POP  ES
    END;

    (* Arrays fr SchreibeZeichen und
       die Punkt-Routinen initialisieren *)
    FOR i := 1 TO 80 DO
      ZSpalte[i]      := (i - 1) + (i - 1) DIV 8;
      ZVerschieben[i] := (i - 1) MOD 8;
    END;
    FOR i := 0 TO 347 DO
      Zeile[i] := 2000H * (i MOD 4) + 90 * (i DIV 4);
    END;
    SEITE := 1;
    IF NOT FontLesen(FontDateiName) THEN
       ROMFontLesen(TRUE);
    END;
  END InitialisiereGrafik;

  PROCEDURE GrafikEin;
  VAR
    i : CARDINAL;
  BEGIN
    OutPort(03B8H, 130);
    FOR i := 1 TO 12 DO
      OutPort(03B4H, i-1);
      OutPort(03B5H, graph[i]);
    END;
    OutPort(03B8H, 138);
  END GrafikEin;

  PROCEDURE GrafikAus;
  VAR
    i : CARDINAL;
  BEGIN
    OutPort(03B8H, 20H);
    FOR i := 1 TO 12 DO
      OutPort(03B4H, i-1);
      OutPort(03B5H, txt[i]);
    END;
    OutPort(03BFH, 1);
    OutPort(03B8H, 28H);
  END GrafikAus;

  PROCEDURE SichtbareSeite(Seite : CARDINAL);
  VAR
    i : CARDINAL;
  BEGIN
    FOR i := 0 TO 1000 DO END;
    OutPort(03B8H, 10 + 128 * Seite);
    FOR i := 0 TO 1000 DO END;
  END SichtbareSeite;

  PROCEDURE AktiveSeite(Seite : CARDINAL);
  BEGIN
    SEITE := Seite;
  END AktiveSeite;

  PROCEDURE SeiteKopieren(Ziel, Quelle : CARDINAL);
  VAR
    Z, Q : CARDINAL;
  BEGIN
    Z := 0B000H + Ziel * 0800H;
    Q := 0B000H + Quelle * 0800H;
    ASM
      PUSH DS
      PUSH ES
      MOV  AX, Z
      MOV  ES, AX
      MOV  AX, Q
      MOV  DS, AX
      XOR  DI, DI
      XOR  SI, SI
      MOV  CX, 4000H
      REPZ MOVSW
      POP  ES
      POP  DS
    END;
  END SeiteKopieren;

  PROCEDURE SeiteLoeschen(Seite : CARDINAL);
  VAR
    seite : CARDINAL;
  BEGIN
    seite := 0B000H + Seite * 0800H;
    ASM
      PUSH ES
      PUSH DI
      MOV  AX, seite
      MOV  ES, AX
      XOR  DI, DI
      MOV  CX, 4000H
      XOR  AX, AX
      REPZ STOSW
      POP  DI
      POP  ES
    END;
  END SeiteLoeschen;

  PROCEDURE BildschirmLoeschen;
  BEGIN
    SeiteLoeschen(SEITE);
  END BildschirmLoeschen;

  PROCEDURE Punkt(X, Y :CARDINAL; Farbe : CARDINAL);
  VAR
    addr   : CARDINAL;
    seg    : CARDINAL;
    bitnum : CARDINAL;
  BEGIN
    seg  := 0B000H + 0800H * SEITE;
    addr := Zeile[Y];
    ASM
      PUSH ES
      MOV BX, seg
      MOV ES, BX             (* ES = seg                  *)
      MOV BX, X
      MOV CL, 3
      SHR BX, CL             (* BX = x shr 3              *)
      MOV DI, addr
      ADD DI, BX             (* DI = BX + addr            *)
      MOV BX, X
      AND BL, 7              (* BL = x and 7              *)
      MOV CL, 7
      SUB CL, BL             (* CL = 7 - BL               *)
      MOV BH, 1
      SHL BH, CL             (* BH = 1 SHL CL             *)
      MOV BL, ES:[DI]        (* Adresse lesen             *)
      MOV CX, Farbe
      TEST CX, 1             (* farbe = 1 ?               *)
      JZ LOESCHEN            (* nein: nach LOESCHEN       *)
      OR BL, BH              (* Bit setzen                *)
      JMP WEITER
    LOESCHEN:
      XOR BH, 255            (* Bitmaske invertieren      *)
      AND BL, BH             (* Bit loeschen              *)
    WEITER:
      MOV ES:[DI], BL        (* Adresse schreiben         *)
      POP ES
    END;
  END Punkt;

  PROCEDURE PunktInvertieren(X, Y : CARDINAL);
  VAR
    addr   : CARDINAL;
    seg    : CARDINAL;
    bitnum : CARDINAL;
  BEGIN
    seg := 0B000H + 0800H * SEITE;
    addr := Zeile[Y];
    ASM
      PUSH ES
      MOV BX, seg
      MOV ES, BX             (* ES = seg                  *)
      MOV BX, X
      MOV CL, 3
      SHR BX, CL             (* BX = x shr 3              *)
      MOV DI, addr
      ADD DI, BX             (* DI = BX + addr            *)
      MOV BX, X
      AND BL, 7              (* BL = x and 7              *)
      MOV CL, 7
      SUB CL, BL             (* CL = 7 - BL               *)
      MOV BH, 1
      SHL BH, CL             (* BH = 1 SHL CL             *)
      MOV BL, ES:[DI]        (* Adresse lesen             *)
      XOR BL, BH             (* Bit invertieren           *)
      MOV ES:[DI], BL        (* Adresse schreiben         *)
      POP ES
    END;
  END PunktInvertieren;

  PROCEDURE PunktTest(X, Y : CARDINAL) : CARDINAL;
  VAR
    addr     : CARDINAL;
    seg      : CARDINAL;
    ergebnis : CARDINAL;
  BEGIN
    seg := 0B000H + 0800H * SEITE;
    addr := Zeile[Y];
    ASM
      PUSH ES
      MOV BX, seg
      MOV ES, BX             (* ES = seg                 *)
      MOV BX, X
      MOV CL, 3
      SHR BX, CL             (* BX = x shr 3             *)
      MOV DI, addr
      ADD DI, BX             (* DI = BX + addr           *)
      MOV BX, X
      AND BL, 7              (* BL = x and 7             *)
      MOV CL, 7
      SUB CL, BL             (* CL = 7 - BL              *)
      MOV BH, 1
      SHL BH, CL             (* BH = 1 SHL CL            *)
      MOV AH, BH
      MOV AL, ES:[DI]        (* Adresse lesen            *)
      MOV BX, 1
      TEST AL, AH            (* Bit = 1 ?                *)
      JNZ WEITER             (* ja: nach WEITER          *)
      MOV BX, 0              (* Bit nicht gesetzt        *)
    WEITER:
      MOV AX, BX
      POP ES
      MOV ergebnis, AX
    END;
    RETURN ergebnis;
  END PunktTest;

  PROCEDURE SchreibeZeichen(Zeichen   : CHAR;
                            XPosition,
                            YPosition,
                            Farbe     : CARDINAL);
  VAR
    Byte        : CHAR;
    Verschieben : CARDINAL;
    Spalte      : CARDINAL;
    i, Basis    : CARDINAL;
    page        : CARDINAL;
  BEGIN
    page := 0B000H + 0800H * SEITE;
    Verschieben := ZVerschieben[XPosition];
    Spalte      := ZSpalte[XPosition];
    FOR i := 0 TO 13 DO
      Byte  := Font[ORD(Zeichen)][i];
      Basis := Zeile[(YPosition-1)*14+i];
      ASM
        MOV  AL,Byte
        XOR  AH,AH
        MOV  BX,Zeichen
        AND  BX,0E0H         (* Auf 9-Bit Breite fr      *)
        CMP  BX,0C0H         (* Grafikzeichen testen      *)
        JNE  Nicht9Bit
        TEST AL,1
        JZ   Nicht9Bit
        MOV  AH,80H          (* 9. Bit setzen             *)
      Nicht9Bit:
        PUSH ES
        MOV  BX,Basis        (* Zeilenadresse in BX       *)
        MOV  DI,Spalte       (* Spaltenoffset in DI       *)
        MOV  CX,page         (* Segment der Seite in ES   *)
        MOV  ES,CX
        MOV  DX,ES:[BX+DI]   (* Inhalt des selektierten   *)
                             (* Wortes auslesen           *)
        PUSH AX
        MOV  AX,0007FH
        MOV  CX,Verschieben
        ROR  AX,CL           (* Maske erzeugen            *)
        XCHG AL,AH
        AND  DX,AX
        POP  AX
        MOV  SI,Farbe
        OR   SI,SI           (* Farbe testen              *)
        JNZ  Normal
                       (* Farbe = 0 : Inverse Darstellung *)
        NOT  AX
        AND  AX,080FFH
                       (* Farbe = 1 : normale Darstellung *)
      Normal:
        XCHG AL,AH
        SHR  AX,CL
        XCHG AL,AH
        OR   AX,DX
        MOV  ES:[BX+DI],AX          (* modifiziertes Wort *)
             (* in den Bildschirmspeicher zurckschreiben *)
        POP  ES
      END;
    END;
  END SchreibeZeichen;

  PROCEDURE SchreibeString(String     : ARRAY OF CHAR;
                           XPosition,
                           YPosition,
                           Farbe      : CARDINAL);
  VAR
    i : CARDINAL;
  BEGIN
    i := 0;
    WHILE (i < Length(String)) DO
       SchreibeZeichen(String[i],
                       XPosition+i, YPosition, Farbe);
      i := i + 1;
    END;
  END SchreibeString;

  PROCEDURE VertLinie(X, Y1, Y2, Farbe : CARDINAL);
  VAR
    addr   : CARDINAL;
    seg    : CARDINAL;
    bitnum : CARDINAL;
    y      : CARDINAL;
    byte   : CARDINAL;
    offset : CARDINAL;
  BEGIN
    seg := 0B000H + 0800H * SEITE;
    ASM
      MOV BX, X
      MOV CL, 3
      SHR BX, CL             (* BX = x shr 3 *)
      MOV offset, BX
      MOV BX, X
      AND BL, 7              (* BL = x and 7 *)
      MOV CL, 7
      SUB CL, BL             (* CL = 7 - BL *)
      MOV BH, 1
      SHL BH, CL             (* BH = 1 SHL CL *)
      MOV byte, BH           (* Zwischenspeichern *)
    END;
    FOR y := Y1 TO Y2 DO
      addr := Zeile[y];
      ASM
        PUSH ES
        MOV BX, seg
        MOV ES, BX           (* ES = seg                  *)
        MOV BX, offset
        MOV DI, addr
        ADD DI, BX           (* DI = BX + addr            *)
        MOV BH, byte         (* Byte lesen                *)
        MOV BL, ES:[DI]      (* Adresse lesen             *)
        MOV CX, Farbe
        TEST CX, 1           (* farbe = 1 ?               *)
        JZ LOESCHEN          (* nein: nach LOESCHEN       *)
        OR BL, BH            (* Bit setzen                *)
        JMP WEITER
      LOESCHEN:
        XOR BH, 255          (* Bitmaske invertieren      *)
        AND BL, BH           (* Bit loeschen              *)
      WEITER:
        MOV ES:[DI], BL      (* Adresse schreiben         *)
        POP ES
      END;
    END;
  END VertLinie;

  PROCEDURE HorLinie(X1, X2, Y,Farbe : CARDINAL);
  VAR
    i, x1, x2, page,
    bit1, bit2,
    byte1, byte2    : CARDINAL;
    zeile           : CARDINAL;
    b1, b2          : CHAR;
  BEGIN
    page := 0B000H + SEITE * 800H;
    zeile := Zeile[Y];
    IF X1 > X2 THEN
      x1 := X2;
      x2 := X1;
    ELSE
      x1 := X1;
      x2 := X2;
    END;
    bit1  := 7 - (x1 MOD 8);
    byte1 := x1 DIV 8;
    bit2  := (x2 MOD 8);
    byte2 := x2 DIV 8;
    ASM
      MOV  CX,bit2
      MOV  AX,0FF80H
      SHR  AX,CL
      MOV  b2,AL
      MOV  CX,bit1
      MOV  AX,01FFH
      SHL  AX,CL
      MOV  b1,AH
    END;
    IF byte1 <> byte2 THEN
      ASM
        PUSH ES
        MOV  AX,page
        MOV  ES,AX
        MOV  BX,zeile
        ADD  BX,byte1
        MOV  AL,b1
        MOV  AH,ES:[BX]
        MOV  DX,Farbe
        OR   DX,DX
        JE   weiss1
        OR   AH,AL
        JMP  weiter1
      weiss1:
        NOT  AL
        AND  AH,AL
      weiter1:
        MOV  ES:[BX],AH
        POP  ES
      END;
      FOR i := (byte1+1) TO (byte2-1) DO
        ASM
          PUSH ES
          MOV  AX,page
          MOV  ES,AX
          MOV  BX,zeile
          ADD  BX,i
          MOV  DX,Farbe
          OR   DX,DX
          JE   weiss2
          MOV  AL,0FFH
          JMP  weiter2
        weiss2:
          XOR  AL,AL
        weiter2:
          MOV  ES:[BX],AL
          POP  ES
        END;
      END;
      ASM
        PUSH ES
        MOV  AX,page
        MOV  ES,AX
        MOV  BX,zeile
        ADD  BX,byte2
        MOV  AL,b2
        MOV  AH,ES:[BX]
        MOV  DX,Farbe
        OR   DX,DX
        JE   weiss3
        OR   AH,AL
        JMP  weiter3
      weiss3:
        NOT  AL
        AND  AH,AL
      weiter3:
        MOV  ES:[BX],AH
        POP  ES
      END;
    ELSE
      ASM
        PUSH ES
        MOV  AX,page
        MOV  ES,AX
        MOV  BX,zeile
        ADD  BX,byte1
        MOV  AL,b1
        MOV  AH,b2
        AND  AL,AH
        MOV  AH,ES:[BX]
        MOV  DX,Farbe
        OR   DX,DX
        JE   weiss4
        OR   AH,AL
        JMP  weiter4
      weiss4:
        NOT  AL
        AND  AH,AL
      weiter4:
        MOV  ES:[BX],AH
        POP  ES
      END;
    END;
  END HorLinie;

END Graf0Hrc.
(* ------------------------------------------------------ *)
(*                Ende von GRA0HRC.MOD                    *)
